/*
 *
 *  oFono - Open Source Telephony
 *
 *  Copyright (C) 2014  Canonical Ltd. All rights reserved.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#ifndef RILMODEM_VENDOR_H
#define RILMODEM_VENDOR_H

enum ofono_ril_vendor {
	OFONO_RIL_VENDOR_AOSP = 0,
	OFONO_RIL_VENDOR_MTK,
	OFONO_RIL_VENDOR_INFINEON,
	OFONO_RIL_VENDOR_QCOM_MSIM,
	OFONO_RIL_VENDOR_IMC_SOFIA3GR
};

#endif /* RILMODEM_VENDOR_H */
