extern struct ofono_plugin_desc __ofono_builtin_udevng;
extern struct ofono_plugin_desc __ofono_builtin_rildev;
extern struct ofono_plugin_desc __ofono_builtin_ril;
extern struct ofono_plugin_desc __ofono_builtin_infineon;
extern struct ofono_plugin_desc __ofono_builtin_ril_intel;
extern struct ofono_plugin_desc __ofono_builtin_rilmodem;
extern struct ofono_plugin_desc __ofono_builtin_isimodem;
extern struct ofono_plugin_desc __ofono_builtin_isiusb;
extern struct ofono_plugin_desc __ofono_builtin_n900;
extern struct ofono_plugin_desc __ofono_builtin_u8500;
extern struct ofono_plugin_desc __ofono_builtin_qmimodem;
extern struct ofono_plugin_desc __ofono_builtin_gobi;
extern struct ofono_plugin_desc __ofono_builtin_mbimmodem;
extern struct ofono_plugin_desc __ofono_builtin_mbim;
extern struct ofono_plugin_desc __ofono_builtin_atmodem;
extern struct ofono_plugin_desc __ofono_builtin_nwmodem;
extern struct ofono_plugin_desc __ofono_builtin_swmodem;
extern struct ofono_plugin_desc __ofono_builtin_ztemodem;
extern struct ofono_plugin_desc __ofono_builtin_iceramodem;
extern struct ofono_plugin_desc __ofono_builtin_huaweimodem;
extern struct ofono_plugin_desc __ofono_builtin_calypsomodem;
extern struct ofono_plugin_desc __ofono_builtin_mbmmodem;
extern struct ofono_plugin_desc __ofono_builtin_telitmodem;
extern struct ofono_plugin_desc __ofono_builtin_hsomodem;
extern struct ofono_plugin_desc __ofono_builtin_ifxmodem;
extern struct ofono_plugin_desc __ofono_builtin_stemodem;
extern struct ofono_plugin_desc __ofono_builtin_dunmodem;
extern struct ofono_plugin_desc __ofono_builtin_hfpmodem;
extern struct ofono_plugin_desc __ofono_builtin_speedupmodem;
extern struct ofono_plugin_desc __ofono_builtin_ubloxmodem;
extern struct ofono_plugin_desc __ofono_builtin_gemaltomodem;
extern struct ofono_plugin_desc __ofono_builtin_xmm7modem;
extern struct ofono_plugin_desc __ofono_builtin_phonesim;
extern struct ofono_plugin_desc __ofono_builtin_cdmamodem;
extern struct ofono_plugin_desc __ofono_builtin_g1;
extern struct ofono_plugin_desc __ofono_builtin_wavecom;
extern struct ofono_plugin_desc __ofono_builtin_calypso;
extern struct ofono_plugin_desc __ofono_builtin_mbm;
extern struct ofono_plugin_desc __ofono_builtin_hso;
extern struct ofono_plugin_desc __ofono_builtin_zte;
extern struct ofono_plugin_desc __ofono_builtin_huawei;
extern struct ofono_plugin_desc __ofono_builtin_sierra;
extern struct ofono_plugin_desc __ofono_builtin_novatel;
extern struct ofono_plugin_desc __ofono_builtin_palmpre;
extern struct ofono_plugin_desc __ofono_builtin_ifx;
extern struct ofono_plugin_desc __ofono_builtin_ste;
extern struct ofono_plugin_desc __ofono_builtin_stemgr;
extern struct ofono_plugin_desc __ofono_builtin_caif;
extern struct ofono_plugin_desc __ofono_builtin_cinterion;
extern struct ofono_plugin_desc __ofono_builtin_gemalto;
extern struct ofono_plugin_desc __ofono_builtin_nokia;
extern struct ofono_plugin_desc __ofono_builtin_nokiacdma;
extern struct ofono_plugin_desc __ofono_builtin_linktop;
extern struct ofono_plugin_desc __ofono_builtin_icera;
extern struct ofono_plugin_desc __ofono_builtin_alcatel;
extern struct ofono_plugin_desc __ofono_builtin_speedup;
extern struct ofono_plugin_desc __ofono_builtin_speedupcdma;
extern struct ofono_plugin_desc __ofono_builtin_samsung;
extern struct ofono_plugin_desc __ofono_builtin_sim900;
extern struct ofono_plugin_desc __ofono_builtin_sim7100;
extern struct ofono_plugin_desc __ofono_builtin_connman;
extern struct ofono_plugin_desc __ofono_builtin_telit;
extern struct ofono_plugin_desc __ofono_builtin_quectel;
extern struct ofono_plugin_desc __ofono_builtin_ublox;
extern struct ofono_plugin_desc __ofono_builtin_xmm7xxx;
extern struct ofono_plugin_desc __ofono_builtin_droid;
extern struct ofono_plugin_desc __ofono_builtin_hfp_bluez5;
extern struct ofono_plugin_desc __ofono_builtin_dun_gw_bluez5;
extern struct ofono_plugin_desc __ofono_builtin_bluez5;
extern struct ofono_plugin_desc __ofono_builtin_hfp_ag_bluez5;
extern struct ofono_plugin_desc __ofono_builtin_upower;
extern struct ofono_plugin_desc __ofono_builtin_provision;
extern struct ofono_plugin_desc __ofono_builtin_cdma_provision;
extern struct ofono_plugin_desc __ofono_builtin_file_provision;
extern struct ofono_plugin_desc __ofono_builtin_smart_messaging;
extern struct ofono_plugin_desc __ofono_builtin_push_notification;
extern struct ofono_plugin_desc __ofono_builtin_allowed_apns;

static struct ofono_plugin_desc *__ofono_builtin[] = {
  &__ofono_builtin_udevng,
  &__ofono_builtin_rildev,
  &__ofono_builtin_ril,
  &__ofono_builtin_infineon,
  &__ofono_builtin_ril_intel,
  &__ofono_builtin_rilmodem,
  &__ofono_builtin_isimodem,
  &__ofono_builtin_isiusb,
  &__ofono_builtin_n900,
  &__ofono_builtin_u8500,
  &__ofono_builtin_qmimodem,
  &__ofono_builtin_gobi,
  &__ofono_builtin_mbimmodem,
  &__ofono_builtin_mbim,
  &__ofono_builtin_atmodem,
  &__ofono_builtin_nwmodem,
  &__ofono_builtin_swmodem,
  &__ofono_builtin_ztemodem,
  &__ofono_builtin_iceramodem,
  &__ofono_builtin_huaweimodem,
  &__ofono_builtin_calypsomodem,
  &__ofono_builtin_mbmmodem,
  &__ofono_builtin_telitmodem,
  &__ofono_builtin_hsomodem,
  &__ofono_builtin_ifxmodem,
  &__ofono_builtin_stemodem,
  &__ofono_builtin_dunmodem,
  &__ofono_builtin_hfpmodem,
  &__ofono_builtin_speedupmodem,
  &__ofono_builtin_ubloxmodem,
  &__ofono_builtin_gemaltomodem,
  &__ofono_builtin_xmm7modem,
  &__ofono_builtin_phonesim,
  &__ofono_builtin_cdmamodem,
  &__ofono_builtin_g1,
  &__ofono_builtin_wavecom,
  &__ofono_builtin_calypso,
  &__ofono_builtin_mbm,
  &__ofono_builtin_hso,
  &__ofono_builtin_zte,
  &__ofono_builtin_huawei,
  &__ofono_builtin_sierra,
  &__ofono_builtin_novatel,
  &__ofono_builtin_palmpre,
  &__ofono_builtin_ifx,
  &__ofono_builtin_ste,
  &__ofono_builtin_stemgr,
  &__ofono_builtin_caif,
  &__ofono_builtin_cinterion,
  &__ofono_builtin_gemalto,
  &__ofono_builtin_nokia,
  &__ofono_builtin_nokiacdma,
  &__ofono_builtin_linktop,
  &__ofono_builtin_icera,
  &__ofono_builtin_alcatel,
  &__ofono_builtin_speedup,
  &__ofono_builtin_speedupcdma,
  &__ofono_builtin_samsung,
  &__ofono_builtin_sim900,
  &__ofono_builtin_sim7100,
  &__ofono_builtin_connman,
  &__ofono_builtin_telit,
  &__ofono_builtin_quectel,
  &__ofono_builtin_ublox,
  &__ofono_builtin_xmm7xxx,
  &__ofono_builtin_droid,
  &__ofono_builtin_hfp_bluez5,
  &__ofono_builtin_dun_gw_bluez5,
  &__ofono_builtin_bluez5,
  &__ofono_builtin_hfp_ag_bluez5,
  &__ofono_builtin_upower,
  &__ofono_builtin_provision,
  &__ofono_builtin_cdma_provision,
  &__ofono_builtin_file_provision,
  &__ofono_builtin_smart_messaging,
  &__ofono_builtin_push_notification,
  &__ofono_builtin_allowed_apns,
  NULL
};
