/*
 *
 *  oFono - Open Source Telephony
 *
 *  Copyright (C) 2008-2011  Intel Corporation. All rights reserved.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#include <drivers/atmodem/atutil.h>

extern void huawei_ussd_init(void);
extern void huawei_ussd_exit(void);

extern void huawei_voicecall_init(void);
extern void huawei_voicecall_exit(void);

extern void huawei_audio_settings_init(void);
extern void huawei_audio_settings_exit(void);

extern void huawei_radio_settings_init(void);
extern void huawei_radio_settings_exit(void);

extern void huawei_gprs_context_init(void);
extern void huawei_gprs_context_exit(void);

extern void huawei_cdma_netreg_init(void);
extern void huawei_cdma_netreg_exit(void);
