/*
 *
 *  oFono - Open Source Telephony
 *
 *  Copyright (C) 2008-2011  Intel Corporation. All rights reserved.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

enum ofono_vendor {
	OFONO_VENDOR_GENERIC = 0,
	OFONO_VENDOR_CALYPSO,
	OFONO_VENDOR_IFX,
	OFONO_VENDOR_STE,
	OFONO_VENDOR_MBM,
	OFONO_VENDOR_GOBI,
	OFONO_VENDOR_QUALCOMM_MSM,
	OFONO_VENDOR_DROID,
	OFONO_VENDOR_OPTION_HSO,
	OFONO_VENDOR_ZTE,
	OFONO_VENDOR_HUAWEI,
	OFONO_VENDOR_SIERRA,
	OFONO_VENDOR_NOVATEL,
	OFONO_VENDOR_WAVECOM,
	OFONO_VENDOR_NOKIA,
	OFONO_VENDOR_PHONESIM,
	OFONO_VENDOR_TELIT,
	OFONO_VENDOR_SPEEDUP,
	OFONO_VENDOR_SAMSUNG,
	OFONO_VENDOR_SIMCOM,
	OFONO_VENDOR_SIMCOM_SIM900,
	OFONO_VENDOR_ICERA,
	OFONO_VENDOR_WAVECOM_Q2XXX,
	OFONO_VENDOR_ALCATEL,
	OFONO_VENDOR_QUECTEL,
	OFONO_VENDOR_QUECTEL_EC2X,
	OFONO_VENDOR_QUECTEL_SERIAL,
	OFONO_VENDOR_UBLOX,
	OFONO_VENDOR_XMM,
	OFONO_VENDOR_GEMALTO,
};
